/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 *
 * $Id: analyze_response_header.c,v 1.9 2000/03/07 15:06:47 goldbach Exp $
 *
 * $Log: analyze_response_header.c,v $
 * Revision 1.9  2000/03/07 15:06:47  goldbach
 * added GPL info
 *
 * Revision 1.8  1999/08/10 17:50:45  gellert
 * fixed a small bug that copied not enough bytes from the
 * HTTP 1.x bla bla  OK line. The \0 at the end was not copied,
 * so there were some of the letters from the previous contents
 * following...
 *
 * Revision 1.7  1999/08/04 11:37:32  gellert
 * Fixed processing of the response line. The reason is not truncated
 * after the first space anymore...
 *
 * Revision 1.6  1999/07/22 14:28:00  goldbach
 * minor cleanup for logmsg() (again)
 *
 * Revision 1.5  1999/07/21 13:15:17  goldbach
 * minor cleanup for logmsg()
 *
 * Revision 1.4  1999/06/09 20:47:42  gellert
 * just some little includes and prototypes against the
 * warnings.
 *
 * Revision 1.3  1999/05/27 16:47:58  goldbach
 * minor cleanup for logmsg()
 *
 * Revision 1.2  1999/05/12 16:04:19  goldbach
 * major clean-up session
 *
 * Revision 1.1  1999/03/30 12:30:06  goldbach
 * 1st surfable version
 *
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <syslog.h>
#include <unistd.h>
#include <string.h>

#include "httpf.h"
#include "mythreads.h"
#include "analyze_header.h"
#include "analyze_response_header.h"
#include "misc.h"

int analyze_response_header(TCT *tct)
{
    int header_start;
    
    /* process response line and get start of optional header */
    header_start = process_http_response_line(tct);
    if (header_start==0)
	return 1;

    logmsg(LOG_DEBUG, "[%d] Optional HTTP header starts at offset %d.",
	   tct->tid, header_start);
    
    /* end of input buffer reached? */
    if (tct->inlen==header_start)
	return 0;

    /* FIXME: fiddle around with the response line */

    /* send response line to output buffer */
    write_response_line(tct);
    
    return process_http_header(tct, header_start);
}



/*
 * process_http_response_line
 * 
 * Examine the first header line and look for version,
 * status code ans the reason phrase.
 * Returns:
 *   0 -- error
 *   >0 -- start of optional header
 */

int process_http_response_line(TCT *tct)
{
    char *eol, *start_reason;
    char reason[LEN_REASON];
    
    /* find next line and replace CRLF with null chars */
    eol=find_eol(tct->inbuf, HTTP_BUFSIZE, 1);
    eol++; /* go to first char of next line */

    /* the reason can contain spaces so we first scan the rest
       and then get the reason */
    sscanf(tct->inbuf, "HTTP/%d.%d %3d", &(tct->version_major), &(tct->version_minor), &(tct->response_status));
    start_reason = strnchr(tct->inbuf, ' ', eol - tct->inbuf);
    start_reason = start_reason + 5; /* skip the space, the number & the space */
    if (start_reason > eol) {
      logmsg(LOG_NOTICE, "[%d] HTTP Header contains no reason: %s", tct->tid, tct->inbuf);
      tct->reason[0] = 0;
    } else {
      if ((strlen(start_reason)+1) < LEN_REASON)
       strncpy(reason, start_reason, strlen(start_reason)+1);
      strncpy(tct->reason, reason, LEN_REASON);
    }
    
    logmsg(LOG_NOTICE,"[%d] Received server response: HTTP/%d.%d, status %d, reason %s.",
	   tct->tid, tct->version_major, tct->version_minor, tct->response_status, tct->reason);
    
    /* optional header starts here */
    return (int)(eol-tct->inbuf);
}

/*
 * write_response_line
 *
 * write version, status and reason
 */
void write_response_line(TCT *tct)
{
    char buf[20];
    
    sprintf(buf, "HTTP/%d.%d %d ", tct->version_major, tct->version_minor, tct->response_status);
    output_string(buf, strlen(buf), tct);

    output_string(tct->reason, strlen(tct->reason), tct);
    output_string("\r\n", 2, tct);    

}

