/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* copy-pkts.c                          el 26.11.92 */
/*
 *
 * $Id: copy-pkts.c,v 1.20 2000/05/07 15:05:04 glauschwuffel Exp $
 * $Log: copy-pkts.c,v $
 * Revision 1.20  2000/05/07 15:05:04  glauschwuffel
 * added init. of context flags end_to_search and more_input_data. added include ftype_check
 *
 * Revision 1.19  2000/03/07 15:06:48  goldbach
 * added GPL info
 *
 * Revision 1.18  2000/03/01 17:43:55  goldbach
 * introduced some constants for read and copy-pkts
 *
 * Revision 1.17  1999/11/30 16:57:52  gellert
 * skip flag: initialization moved to beginning of copy-pakets. read_header: find_empty_line is called with the length of the read block, not the full HTTP_BUFSIZE.
 *
 * Revision 1.16  1999/11/11 16:50:47  gellert
 * Changed a log message.
 *
 * Revision 1.15  1999/11/11 15:50:33  gellert
 * Changed the processing (= reading) of client requests.
 * Now the bodies are copied as well and "POST" works.
 *
 * Revision 1.14  1999/09/09 15:40:25  gellert
 * if there is no body following directly after the header
 * (in the first datagram) then the proxy now tries to get
 * the next datagram. because squid often splits the packets
 * and sends header in body in different datagrams.
 *
 * Revision 1.13  1999/08/04 12:00:52  gellert
 * Adapted to the change of typecheck (tct->mime_type is no a *char).
 * When an empty body arrives from the server the header is send
 * immediately to the client and that's it. In previous versions
 * the type check failed for all cached gifs etc, because the body
 * of the server response has size 0 (in the header is the response
 * line: 304 Not Modified).
 *
 * Revision 1.12  1999/07/22 15:35:03  goldbach
 * yet again cleanup for logmsg()
 *
 * Revision 1.11  1999/07/22 14:28:00  goldbach
 * minor cleanup for logmsg() (again)
 *
 * Revision 1.10  1999/07/15 16:55:17  gellert
 * Ok, nun inverse Logik. Wenn Mimetype NICHT html, dann
 * durch den Typecheck. Sonst filtern. Vielleicht noch mal
 * testen, was bei Text passiert!!!
 *
 * Revision 1.9  1999/06/09 20:33:52  gellert
 * new version. now the parsing of tags and attributes works.
 * and it still works if the client cancels the transmission
 * (stop button in browser).
 *
 * Revision 1.8  1999/05/27 16:47:59  goldbach
 * minor cleanup for logmsg()
 *
 * Revision 1.7  1999/05/27 11:29:34  gellert
 * well, now it does what we want. just small mistakes (eg end of
 * comments) are done.
 *
 * Revision 1.6  1999/05/12 16:04:20  goldbach
 * major clean-up session
 *
 * Revision 1.5  1999/03/30 12:30:07  goldbach
 * 1st surfable version
 *
 * Revision 1.4  1999/02/16 17:43:56  gellert
 * now it works. up to now the header is read and passed on to the
 * server, the other direction is just copied through.
 *
 * Revision 1.3  1999/02/11 14:38:48  gellert
 * ein paar alte defines ausgemerzt, neue kommentare fuer kommende
 * funktionen schon mal eingefuegt.
 *
 * Revision 1.2  1999/02/09 13:39:28  gellert
 * removed all the "ifdef ssl/des3/arc4" - parts, so the code is now
 * a little bit simpler & shorter :-)
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <syslog.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "mythreads.h"
#include "read.h"
#include "analyze_header.h"
#include "analyze_request_header.h"
#include "analyze_response_header.h"
#include "filter_html.h"
#include "misc.h"
#include "copy-pkts.h"
#include "ftype_check.h"

static int process_client(TCT *);
static int process_server(TCT *);

int copy_pkts (TCT *tct) {
  int 	nfds = getdtablesize(); /* select ...? */
  fd_set 	fds;	/* fuer select: liste mit sockin + sockout */
  static struct timeval	tout;
  int result;
  
  FD_ZERO(&fds);
  
  while (1) {
      tout.tv_sec = 60*60*2;
      tout.tv_usec = 0;

      FD_SET(tct->insocket, &fds);
      FD_SET(tct->outsocket, &fds);

      /*
       * reset skip flag. we reset it here regardless of wether select()
       * returns with a call from the client or the server.
       * Since the http-header functions use output_string() and
       * the skip_flag is used in there we reset it here.
       */
      tct->skip_flag=0;

	  /*
	   * reset flag telling if we're looking for the end of comment
       * or end of tag.
	   */
	  tct->end_to_search=END_OF_TAG;

      if (select(nfds, &fds, NULL, NULL, &tout) > 0) {
	  /* request? */
	  if (FD_ISSET(tct->insocket, &fds)) {
	      result = process_client(tct);
	      if (result != PROCESS_REQUEST_OK) {
		logmsg(LOG_DEBUG, "[%d] process_client failed.", tct->tid);
	        return result;
	      }
	  }

	  /* response? */
	  if (FD_ISSET(tct->outsocket, &fds)) {
	      result = process_server(tct);
	      if (result < PROCESS_RESPONSE_OK) {/* severe error */
		logmsg(LOG_DEBUG, "[%d] process_server failed.", tct->tid);
	        return result;
	      }
	      break; /* eof/everything ok: next select */
	  }
      } else {
	logmsg(LOG_ERR,
	       "[%d] Select on server and client sockets failed with errno %d.",
	       tct->tid, errno);
      }
  }
  
  return 0; 
}

/*
 * process_client
 *
 * receive the header of the http-request. parse it
 * and modify it if necessary. then send it.
 */

static int process_client(TCT *tct) {
  int result;
    
  reset_buffers(tct);   /* reset buffer offsets */
    
  switch(read_client_header(tct)) {
  case READ_SOCKET_ERROR:
  case READ_EOF:
	return PROCESS_REQUEST_FAILED;
  case READ_NO_END_OF_HEADER: /* we have no header */
	logmsg(LOG_ERR, "[%d] Client request contains no header.", tct->tid);
	logmsg(LOG_DEBUG, "[%d] Inbuf offset is %d of %d bytes.",
	       tct->tid, tct->inoffset, tct->inlen);
	
	return PROCESS_REQUEST_OK;
	
	break;
  default: /* READ_OK */
	/*
	 * Reading succeeded. Now analyze the inbuf and fill
	 * the outbuffer.
	 */
	result = analyze_request_header(tct);

	/* now the (new generated) header is in the output buffer */
	
	/* and now process the body if it exists. */
	if (tct->inoffset<tct->inlen) {
	  /* until now the client body is just copied! */
	  logmsg(LOG_ERR, "[%d] Notice: Client bodies are just copied.", tct->tid);
	  output_string(tct->inbuf+tct->inoffset, tct->inlen-tct->inoffset,
					tct);
	} else {
	  logmsg(LOG_DEBUG, "[%d] Received empty body on client socket.", tct->tid);
	}
	
	/* send buffer */
	logmsg(LOG_INFO, "[%d] Sending client header containing %d bytes.",
		   tct->tid, tct->outoffset);
	if (send(tct->outsocket, tct->outbuf, tct->outoffset, 0) != tct->outoffset) {
	  logmsg(LOG_ERR, "[%d] Write error on outsocket %d with errno.",
			 tct->tid, tct->outsocket);
	  return PROCESS_REQUEST_FAILED;
	}
	
	break;
  } /* switch */
  
  return PROCESS_REQUEST_OK;
}


/*
 * process_server
 *
 * receive the header of the http-response. parse it
 * and modify it if necessary. then send it.
 */

static int process_server(TCT *tct) 
{
    int result;
    int quit=0; /* for the read-body loop */
    int at_beginning_of_body=0; /* used for checking body contents */

    /* reset mime type */
    tct->mime_type[0] = '\0';
    
	/* reset flag telling that we get more input */
	tct->more_input_data=YES;

    /* reset method: servers don't have methods so a value of 0 indicates server processing FIXME: this should be a contstant :) */
    tct->method = 0;
    
    reset_buffers(tct);   /* reset buffer offsets */
    result = read_server_header(tct);    /* read header */

    switch (result) {
    case READ_NO_END_OF_HEADER:
	logmsg(LOG_ERR, "[%d] FIXME: We should generate an error page here...", tct->tid);
	logmsg(LOG_ERR, "[%d] No status line in server response!", tct->tid);
	return PROCESS_RESPONSE_OK;
	
    case READ_EOF:
	logmsg(LOG_ERR, "[%d] FIXME: We should generate an error page here...", tct->tid);
	logmsg(LOG_ERR, "[%d] EOF on server socket.", tct->tid);
	return PROCESS_RESPONSE_OK;
	
    case READ_SOCKET_ERROR:
	logmsg(LOG_ERR, "[%d] FIXME: We should generate an error page here...", tct->tid);
	logmsg(LOG_ERR, "[%d] Receiving data on server socket failed with errno %d.",
	       tct->tid, errno);
	return PROCESS_RESPONSE_OK;
	
    default:
	logmsg(LOG_DEBUG, "[%d] Analyzing response header.", tct->tid);
	logmsg(LOG_DEBUG, "[%d] Am now at %d of %d bytes in inbuf",
	       tct->tid, tct->inoffset, tct->inlen);
	
	analyze_response_header(tct);
	at_beginning_of_body=1; /* we are at the beginning */

	/* outbuf is filled with the header we want to send */
	
	break;
    }

    logmsg(LOG_DEBUG, "[%d] Done with header anlysis", tct->tid);
    logmsg(LOG_DEBUG, "[%d] Am now at %d of %d bytes in inbuf.",
	   tct->tid, tct->inoffset, tct->inlen);

    /* read data as long as we get some */
    if (tct->inoffset == tct->inlen) {
      /* ok, there is no body */
      logmsg(LOG_DEBUG, "[%d] No document body.", tct->tid);
      if (send(tct->insocket, tct->outbuf, tct->outoffset, 0)<0) {
	logmsg(LOG_ERR, "[%d] Error on sending server header without body.", tct->tid);
	return PROCESS_RESPONSE_OK;
	}

      /* is the body following in the next packet? */
      reset_buffers(tct);
      if (read_server_body(tct) != READ_OK) {
	/* leave if reading failed */
	logmsg(LOG_INFO, "[%d] There is no body following after the header.",
        	tct->tid);
        result = READ_EOF;	/* this means we leave the do loop immediately */
      } else {
        logmsg(LOG_INFO, "[%d] Got new datagram with body.", tct->tid);
      }

    }
    	                                    

    quit=0;
    
    do {
	switch(result) {
	case READ_SOCKET_ERROR:
	    logmsg(LOG_ERR, "[%d] FIXME: We should generate an error page here...", tct->tid);
	    logmsg(LOG_ERR, "[%d] Receiving data on server socket failed with errno %d.",
		   tct->tid, errno);
	    return PROCESS_RESPONSE_FAILED;
	case READ_EOF:
	    quit=1; /* leave on eof */
	    break;
	    	    
	default:
	    /* if there's html: filter it! */
	    /* logmsg(LOG_DEBUG, "[%d] Mime-Type: %s", tct->tid, tct->mime_type); */
	    if (strncmp(tct->mime_type, "text/html", min(strlen(tct->mime_type), 9))!=0) {
		/* is this the first datagram of this connection? */
		if (at_beginning_of_body!=0) {
		    at_beginning_of_body=0;
		    logmsg(LOG_DEBUG, "[%d] Checking MIME-type at inbuf offset %d.",
			   tct->tid, tct->inoffset);
		    
		    /* check the body we receive */
		    /* we only let things pass that we know (eg. gifs, jpegs) */
		
		    if(type_check(tct)!=0)
		    {
			logmsg(LOG_ERR, "[%d] FIXME: Illegal non-HTML data", tct->tid);
			logmsg(LOG_ERR, "[%d] (this should be an error page)", tct->tid);
			return PROCESS_RESPONSE_OK;
		    /* ok, we recognized the data format, so we send it */
		    } else {
			logmsg(LOG_DEBUG, "[%d] Non-HTML Data: Copying %d of %d bytes",
			tct->tid, tct->inlen-tct->inoffset, tct->inlen);
    			output_string(tct->inbuf+tct->inoffset , tct->inlen-tct->inoffset, tct);
			logmsg(LOG_INFO, "[%d] Non-HTML Data: Sending %d bytes",
					tct->tid, tct->outlen);
			if (send(tct->insocket, tct->outbuf, tct->outlen, 0)<0) {
			  return PROCESS_RESPONSE_OK;
			}

			/* FIXME: check if sending succeeded */
			reset_buffers(tct);
		    }

		/* ok, this is not the first datagram of the connection
		   and it's not html. so we just let it pass our proxy */
		} else {
		    
		    if (tct->inlen > 0) {
			/* inbuf und inlen in out aendern, wenn process_body geht */
			logmsg(LOG_INFO, "[%d] Sending server body containing %d bytes.",
			       tct->tid, tct->inlen);
		
			logmsg(LOG_INFO, "[%d] Sending non-HTML data.", tct->tid);
			
			logmsg(LOG_INFO, "[%d] This thread's MIME type is %s.",
			       tct->tid, tct->mime_type);
			/* if I call test_write here I get still a SIGPIPE when I write
			   I don't know why. But I don't get SIGPIPE, when I am just
			   sending and quitting on error... WHY? */
			if (send(tct->insocket, tct->inbuf, tct->inlen, 0) < 0) {
			    logmsg(LOG_INFO, "[%d] Sending on client socket failed with errno %d.",
				   tct->tid, errno);
			    quit = 1;
			}

			/* FIXME: check if sending succeeded */
			reset_buffers(tct);
			
		    }
		} /* else */
                
	    } else {
		if (filter_html(tct) == -1) {
		  quit = 1;
		  }
		
		/* FIXME: check if sending succeeded */
		    
		
	    }
	    
	    result=read_server_body(tct); /* read next block */
	    
	    break;
	    
	}
    } while (quit==0);

    logmsg(LOG_INFO,"[%d] Connection closed by server.", tct->tid);
    
    return PROCESS_RESPONSE_OK;
}

