/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* httpf.c                                         og 09.02.99 */
/* was before: proxy.c
 * previous development see MGproxy, up to version 1.10
 *
 * $Id: httpf.c,v 1.27 2000/05/07 15:19:37 glauschwuffel Exp $
 * $Log: httpf.c,v $
 * Revision 1.27  2000/05/07 15:19:37  glauschwuffel
 * added global var httpf_id which holds a nice HTML comment telling that the document was processed by our proxy. acdded inclusion of cfglists.h. changed inclusion of strings.h to string.h. made the solaris-only bit a local block whithin span_threads.
 *
 * Revision 1.26  2000/04/24 21:22:47  gellert
 * fiddled around with inet address. now client_allowed works on linux
 * and solaris.
 *
 * Revision 1.25  2000/04/24 19:44:34  gellert
 * Added new feature: clients entry in httpf.cfg. Only configured
 * clients are allowed to connect to httpf.
 * The handling of the lists of protocols and the lists
 * of allowed clients are now handled in cfglists.c
 *
 * Revision 1.24  2000/03/07 15:18:00  gellert
 * some changes to port to linux!!!
 *
 * Revision 1.23  2000/03/07 15:06:51  goldbach
 * added GPL info
 *
 * Revision 1.22  2000/03/01 13:33:17  gellert
 * Added a new commandline option "-n" (nodaemon), so the proxy stays
 * in foreground for debugging.
 *
 * Revision 1.21  2000/02/23 15:42:55  goldbach
 * fixed sem_init() call in httpf.c
 *
 * Revision 1.20  2000/02/23 15:09:26  goldbach
 * now we use POSIX-semaphores instead of Solaris-semaphores
 *
 * Revision 1.19  1999/11/26 10:47:07  goldbach
 * New actions introduced: replskip and replendskip skip sections of documents, i.e. data won\'t be sent to the client.
 *
 * Revision 1.18  1999/08/04 11:41:44  gellert
 * Changed the type check. Now there are two kinds of entries in httpf.cfg:
 * 1) ftype mimetype offset pattern: If the given mimetype matches, than
 *    it is checked whether the block contains pattern at offset.
 * 2) ftype mimetype: If the given mimetype matches, no further check follows.
 *    This makes sense for types like text/plain...
 * These changes required the change of the threadcontext, there is now a
 * character string containing the given mimetype of the server response.
 *
 * Revision 1.17  1999/07/28 14:20:07  gellert
 * some changes to the calls of logmsg
 *
 * Revision 1.16  1999/07/22 14:28:01  goldbach
 * minor cleanup for logmsg() (again)
 *
 * Revision 1.15  1999/07/22 13:24:16  goldbach
 * minor cleanup for logmsg()
 *
 * Revision 1.14  1999/07/14 09:50:04  goldbach
 * removed last strncpy.
 *
 * Revision 1.13  1999/07/14 08:42:04  gellert
 * inserted htons to convert port number to network byte order
 *
 * Revision 1.12  1999/06/29 12:21:30  gellert
 * now the debug level is set to (not incremented by) the given
 * command line argument.
 *
 * Revision 1.11  1999/06/23 16:59:53  gellert
 * now the socket option SO_REUSEADDR is set.
 * this prevents us from choosing a new port
 * after every test...
 *
 * Revision 1.10  1999/06/23 10:58:23  gellert
 * ok, inluded setting of reslimit
 *
 * Revision 1.9  1999/06/09 20:33:53  gellert
 * new version. now the parsing of tags and attributes works.
 * and it still works if the client cancels the transmission
 * (stop button in browser).
 *
 * Revision 1.8  1999/05/27 16:47:59  goldbach
 * minor cleanup for logmsg()
 *
 * Revision 1.7  1999/05/27 11:29:35  gellert
 * well, now it does what we want. just small mistakes (eg end of
 * comments) are done.
 *
 * Revision 1.6  1999/05/12 16:04:21  goldbach
 * major clean-up session
 *
 * Revision 1.5  1999/04/28 09:18:40  gellert
 * added the modules to read in the config file and place the
 * tags and attributes into a hash table.
 *
 * Revision 1.4  1999/03/02 11:28:08  goldbach
 * added buffer offsets initialization
 *
 * Revision 1.3  1999/02/16 17:44:34  gellert
 * some adaptions, eg new parameter -v etc.
 *
 * Revision 1.2  1999/02/09 13:39:30  gellert
 * removed all the "ifdef ssl/des3/arc4" - parts, so the code is now
 * a little bit simpler & shorter :-)
 *
 * Revision 1.1.1.1  1999/02/09 13:20:07  gellert
 * Import of httpf
 *
 */

/*
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
*/

#include <sys/param.h>
#include <sys/resource.h>

/*
#include <sys/wait.h>
#include <poll.h>
*/

#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>
/* #include <thread.h> this should not be necessary */
/* #include <synch.h> this should not be necessary */
#include <semaphore.h>
#include "httpf.h"
#include "mythreads.h"
#include "misc.h"
#include "copy-pkts.h"
#include "open_conn.h"
#include "taghash.h"
#include "read_config.h"
#include "cfglists.h"


#define DEFCOPYSIZE 8192
#define DEFSVRPORT 1080
#define DEFAULT_THREADS 50 /* per Default mit 50 Threads gleichzeitig loslegen */
#define	MAXPACKETLEN	65536
#define DEFAULT_LOG_LEVEL LOG_NOTICE


int log_level = DEFAULT_LOG_LEVEL; /* report all errors, criticals, alerts and emergencies */
int real_proxy =1;		/* per default start as real proxy, this gets
			   changed by the -d argument */

struct sockaddr_in	dest_fix_addr; /* Addresse des Zielservers, wenn kein Socks
					* genutzt wird. */

sem_t log_sema; /* semaphore for logging, see logmsg() */

char *httpf_id = "<!-- This page was frobnicated by httpf $Revision: 1.27 $.\nSee http://httpf.sourceforge.net for a description of this security proxy. -->\n\0";

int tune_res_lim ( int threads ) {
#define FDUSED 5 /* listen, log, door, 2 im Sinn ;) */
  struct rlimit maxfiles;
  int files_needed;
  
  getrlimit (RLIMIT_NOFILE, &maxfiles);
  files_needed = 2 * threads + FDUSED;
  
  /*  logmsg( LOG_INFO, "Default-Filelimit: %d Max: %d\n", maxfiles.rlim_cur, maxfiles.rlim_max); */
  
  if (files_needed > maxfiles.rlim_max) {
      logmsg(LOG_WARNING, 
	      "Warning: number of threads (%d) exceeds number of available file descriptors (%d).", 
	      threads,  maxfiles.rlim_max );
      maxfiles.rlim_cur = maxfiles.rlim_max;
  } else if (files_needed > maxfiles.rlim_cur) {
      maxfiles.rlim_cur = files_needed;
  }
  
  if (setrlimit (RLIMIT_NOFILE, &maxfiles)) {
      logmsg( LOG_ERR,
	      "Error: unable to adjust the number of filedescriptors.");
      return (-1);
  }
  
  return( (int) ((maxfiles.rlim_cur - FDUSED) / 2) );
}


int setsocksize(int socket, int size) {
  if(setsockopt(socket, SOL_SOCKET, SO_SNDBUF,
		(char *)&size, sizeof(int)) < 0) {
      logmsg(LOG_ERR, "Setting SNDBUF for socket %d failed.", socket);
      return(-1);
  }
  
  if(setsockopt(socket, SOL_SOCKET, SO_RCVBUF,
		(char *)&size, sizeof(int)) < 0)  {
      logmsg(LOG_ERR, "Setting RCVBUF for socket %d failed.", socket);
      return(-1);
  }
  return(0);
}

int setsockreuse( int socket) {
  int on=1;

  if(setsockopt(socket, SOL_SOCKET, SO_REUSEADDR,
     (char *) &on, sizeof (on)) < 0) {
	logmsg(LOG_ERR, "Setting REUSEADDR for socket %d failed.", socket);
	return(-1);
    }
  
  return(0);
}



/* Anfrage beantworten: ist die Hauptfunktion aller Threads */
/* Die Threads identifizieren sich durch ihren TCT selbst, sodass 
 * Messwerte in diesem Struct gespeichert werden koennen.
 */
int process_query (TCT *context) {
    /* child process */
    struct sockaddr_in	dest_addr;
    
    bcopy((char *) &dest_fix_addr, (char *) &dest_addr, sizeof(dest_addr));
    if (((*context).outsocket = open_conn(&dest_addr, (*context).ServerSockSize,context)) < 0) {
	logmsg(LOG_ERR, "[%d] Can't connect to destination.", context->tid);
	close((*context).insocket);
	return(-1);
    } else {
	logmsg(LOG_NOTICE, "[%d] Successful connection to destination on socket %d.",
	       context->tid, (*context).outsocket);
    }
  
  /* process the request */
  if (copy_pkts(context) < 0) {
      logmsg(LOG_ERR, "[%d] Processing request failed.", context->tid);
      close((*context).insocket);
      close((*context).outsocket);
      return(-1);
  } 
  
  logmsg(LOG_NOTICE, "[%d] Request to %s:%d processed successfully.",
	 context->tid,
	 inet_ntoa(dest_addr.sin_addr),
	 dest_addr.sin_port);
  
  if ((close((*context).insocket))< 0)
    logmsg(LOG_ERR, "[%d] Closing client socket failed.", context->tid);
  if ((close((*context).outsocket)) <0)
    logmsg(LOG_ERR, "[%d] Closing server socket failed.", context->tid);

  return 0;

}


#define SELF (*mycontext)
/* Threaddy ist die Initialfunktion eines Threads, der die Kontrolle vom
 * Kernel nach dem Initialisieren eines neuen Threads gegeben wird.
 *
 * Es gibt vier Zustaende in denen sich ein Thread auf der Anwendungsebene
 * befinden kann:
 *
 * READY:       der Thread ist bereit einen (neuen) Auftrag zu bearbeiten.
 * SCHEDULED:   der Thread hat eine neuen Auftrag bekommen.
 * WORKING:     der Thread bearbeitet einen Auftrag.
 * ERROR:       der Thread hat einen Fehler erkannt oder wurde von einem
 *              anderen Thread als fehlerhaft erkannt.
 */
void threaddy (TCT *mycontext) {
  /* og, 11.02.99: drei puffer anlegen, und nicht malloc benutzen. */
  char inbuffer[HTTP_BUFSIZE];
  char outbuffer[HTTP_BUFSIZE];
  char errorbuffer[HTTP_BUFSIZE];
  char mime_type[LEN_MIMETYPE];

  SELF.inbuf = &inbuffer[0];
  SELF.outbuf = &outbuffer[0];
  SELF.errbuf = &errorbuffer[0];
  SELF.mime_type = &mime_type[0];

  logmsg(LOG_DEBUG, "%s", SELF.mime_type);

  /* gg 02.03.99: initialize offsets */
  SELF.inoffset=0;
  SELF.outoffset=0;
  SELF.erroffset=0;

  /* don't skip document parts */
  SELF.skip_flag=0;
  
  while (1) {
      /* repeat until father dies */
      if (sem_wait (&SELF.wait)) {
	  /* gentle touchdown in thread nirvana */
	  logmsg(LOG_ERR, "[%d] sem_wait Error: pthread_exit\n", mycontext->tid);
	  SELF.stat = ERROR;
	  pthread_exit ((void *)NULL);
      }
      SELF.stat = WORKING;
      
      process_query(mycontext);
      SELF.stat = READY;
    } /* while */
}



/* Versucht die angegebene Anzahl an Threads zu starten. Gibt die Anzahl der
 * tatsaechlich gestarteten Threads zurueck.
 */
int span_threads (int threads, TCT *tcontext) {
  int i, anz;
  pthread_attr_t tattr;
  pthread_t tid;
  
  pthread_attr_init(&tattr);
  /* wichtig, damit Betriebsmittel von Threads automatisch nach der 
   * Termination freigegeben werden. Also beispielsweise wenn sich ein
   * Thread aufgrund eines Fehlers selber terminiert.
   */
  pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED);
  
  for (i = 0, anz = 0; i<threads; i++) {
	if (pthread_create(&tid, &tattr, (void *(*)())threaddy, (void *)(tcontext + i)) == 0) {
	  /* wieder ein neuer Thread erzeugt */
	  anz++;
	} else {
	  /* diesen Thread konnten wir nicht starten, er sollte also
	   * auch keine Auftraege bekommen...
	   */
	  (*(tcontext + i)).stat = ERROR;
	}
  } /* for */
  
  /* dem Scheduler unsere Vorstellungen zur Anzahl der gleichzeitig
   * laufenden Threads mitteilen., cb. 01.12.97
   */
#ifdef solaris
  {
	int ret = thr_setconcurrency(anz);
	if (ret) {
	  /* Fehlerbehandlung: todo*/
	};
  }
#endif
  
  return anz;
}


/* einen Thread finden, der gerade nichts zu tun hat:
 * Rueckgabe => 0: Slot von tcontext mit freiem Thread
 * Rueckgabe = -1: zur Zeit keit freier Tread
 * Rueckgabe < -1: Alle Threads haben terminiert => time to panic
 */
#define TCTP (*(tcontext + i))
int find_thread (int threads, TCT *tcontext) {
    int tid = -1;
    int i, errnr = 0;
    
    for (i=0; i < threads; i++) {
      	logmsg(LOG_NOTICE, "Thread #%d has status %d.", TCTP.tid, TCTP.stat);
	
	if (TCTP.stat == READY) {
	  /* tja, wer Spass haben moechte, laesst diese Anweisung weg */
	    TCTP.stat = SCHEDULED;
	    tid = i;
	    break;
	} else if (TCTP.stat == ERROR) {
	    errnr++;
	}      
    }
  
    if ( (tid == -1) && (errnr == threads) )
	/* alle meine Threads sind terminiert ! */
	return -2;
    else
	return (tid);
}



int main (int argc, char *argv[]) {

  int			sockfd, newsockfd, clilen, c, ret;
  int 			serv_port = DEFSVRPORT;
  int 			threads = DEFAULT_THREADS;
  TCT 			*tcontext;
  struct sockaddr_in	cli_addr, serv_addr;
  struct hostent		*hp;
  char 			dest_host[MAXHOSTNAMELEN];
  char			config_file_name[MAX_STRING_LENGTH];
  struct rlimit		maxfiles;
  int resflag = 0; 	/* flag for child: run setrlimit after forking */

  
  int NoDaemon = 0;
  int ServerSockSize = 0; 
  int ClientSockSize = 0; 
  int CopySize = DEFCOPYSIZE;

  /* semaphore init */
/*  sem_init(&log_sema, 1, USYNC_THREAD, NULL); */
  sem_init(&log_sema, 0, 1);
  
    
  openlog("httpf", LOG_PID, LOG_LOCAL6);
  
  while ((c= getopt(argc, argv, 
		    "?hp:m:ns:S:t:d:D:v:"
		    )) != EOF)
    {
      switch (c) 
	{
	case '?':
	case 'h':
	default:
	  usage();
	  exit(1);
	case 'v': /* change level of verbosity */
		  log_level = atoi(optarg);
		  break;
	case 'p': /* Port an dem der Proxy "listen" soll */
	  serv_port=atoi(optarg);
	  break;
	case 'm': /* Read/Write-Laenge beim Kopieren */
	  CopySize = atoi(optarg);
	  break;
	case 'n': /* No Daemon, stay in foreground... */
	  NoDaemon = 1;
	  break;
	case 's': /* Client-Side Socket Groesse */
	  ClientSockSize = atoi(optarg);
	  break;
	case 'S': /* Server-Side Socket Groesse */
	  ServerSockSize = atoi(optarg);
	  break;
	case 't': /* Anzahl der Threads */
	  threads = atoi(optarg);

	  if ( threads >= 0 ) {
#define FO 7 /* stdin, stdout, stderr, logfile, urlfile, 2 im Sinn ;) */
	      int files_needed;
	      
	      getrlimit (RLIMIT_NOFILE, &maxfiles);
	      files_needed = threads + FO;
	      if (files_needed > maxfiles.rlim_max) {
		  logmsg(LOG_WARNING, "Number of threads exceeds available file descriptors.");
		  logmsg(LOG_WARNING, "Limiting threads to %lu.", maxfiles.rlim_max - FO);
		  threads = (maxfiles.rlim_max - FO);
		  maxfiles.rlim_cur = maxfiles.rlim_max;
		  resflag = 1;
		} else if (files_needed > maxfiles.rlim_cur)
		{
		    maxfiles.rlim_cur = files_needed;
		    resflag = 1;
		}
	      /* An dieser Stelle soll eigentlich der Bedarf an Filehandles
	       * fuer diesen Prozess und alle seine Kindprozesse gesetzt 
	       * werden (pro Thread brauch man mindestens einen Socket, 
	       * zuzueglich ein paar Handles fuer die Standardfiles und
	       * fuer das Logfile.
	       *
	       * Leider setzt der Aufruf von 'Set Session ID' (setsid()) 
	       * im Hauptprogramm unter Solaris 2.5.1 diese Einstellungen 
	       * zurueck. Mann muss diesen Aufruef (setrlimit()) also nach
	       * setsid() wiederholen.
	       * 
	       * Da ich nicht weiss, ob dies ein Bug oder ein Feature von
	       * Solaris ist, lasse ich die Einstellungen trotzdem hier, da
	       * dies die geeignete Stelle zu sein scheint.
	       */
	      if (setrlimit (RLIMIT_NOFILE, &maxfiles))	{
		  logmsg(LOG_ERR, "Unable to adjust the number of file descriptors.");
		  perror ("setrlimit(RLIMIT_NOFILE, ...)");
		  exit (1);
	      }
	    } else 
		return (1);
	  break;
	  
	  break;
	case 'd': /* Zielhost Adresse, wenn Verbindungen immer nur zu 
		   * einem Zielrechner erlaubt werden (wie plug-gw) 
		   */
	  real_proxy = 0;
	  strncpy(dest_host, optarg, sizeof(dest_host) );
	  if ((dest_fix_addr.sin_addr.s_addr = inet_addr(dest_host)) == -1) {
	      /* vielleicht ein DNS-Name? */
	      if ((hp = gethostbyname(dest_host)) == NULL) {
		  logmsg(LOG_ERR, "Can't resolve destination host.");
		  exit(1);
	      } else {
		  /* DNS Adresse */
		  bcopy(hp->h_addr,
			(char *)&dest_fix_addr.sin_addr,
			hp->h_length);
		  dest_fix_addr.sin_family = hp->h_addrtype;
	      }
	  } else {
	      /* IP Adresse */
	      dest_fix_addr.sin_family = AF_INET;
	  }
	  break;
	case 'D': /* Port an den auf dem Zielhost verbunden werden soll */
	  dest_fix_addr.sin_port = htons(atoi(optarg));
	  break;
	}
    }
  
  logmsg(LOG_INFO, "Socket sizes are: server %d, client %d, copy %d.",ServerSockSize, ClientSockSize, CopySize);

  /* get the name of the config file (= name of httpf plus ".cfg") */
  strncpy(config_file_name, argv[0], MAX_STRING_LENGTH-3);
  strncat(config_file_name, ".cfg", 4);
  read_config(config_file_name);
  

   if (NoDaemon==0) {
     if (mkdaemon()) { 
        logmsg(LOG_ERR, "Unable to create daemon process."); 
        exit(-1); 
     } 
   }

  setsid();   
  /* leider setzt setsid() die limits zurueck */
  if (resflag) {
      /* falls die Filehandles angepasst wurden, dann hier nochmal... */ 
      if (setrlimit (RLIMIT_NOFILE, &maxfiles)) {
	  logmsg(LOG_ERR, "Unable to adjust the number of file descriptors.");
	  perror ("setrlimit(RLIMIT_NOFILE, ...)");
	  exit (1);
      } 
  }

/*
 * FIXME: close stdin, stdout, stderr if this should be a real demon
 */
  close (0); close (1); close (2);
  
  /* Keine Zombies bitte */
#ifdef FREEBSD
  signal(SIGCHLD, SIG_IGN);
#else
  signal(SIGCLD, SIG_IGN);
#endif

  /* and no killing if we get SIGPIPE */
  /* this happens if we sent something to the client but he
     canceled the transfer (= stop button) */
  signal(SIGPIPE, SIG_IGN);
   
  /* Ressource-Limits fuer File-Descriptoren (Socket!) hochsetzen.
   * darf erst nach mkdaemon geschehen, da setsid Ressourcen wieder auf Default setzt.
   */
  if ((ret = tune_res_lim( threads )) < threads) { 
      logmsg (LOG_ERR, "Limiting number of threads to %d.", ret);
      threads = ret;
  }
  
  /*********** Anfang Thread-inits */
  
  /* Thread Kontexte allozieren und Vorbelegen */
  
  if ( threads < 1) {
      logmsg (LOG_ERR, "Not enough threads, aborting.");
      exit (1);
  }
  
  if ( (tcontext = (TCT *) malloc (threads * sizeof (TCT)) ) == (TCT *)NULL) {
      logmsg (LOG_ERR, "Not enough memory for malloc.");
      exit (1);
  } else {
      /* Hier die Kontextinformationen eines Threads initialisieren.
       * KONTEXT ist nicht ganz der richtige Ausdruck. Die Kontextinformationen
       * der Threads wird vom Kernel verwaltet. Wir speichern in dem Struct
       * demnach anwendungsbezogene Kontextinformationen,
       */
      int i;
      for (i=0; i<threads; i++)	{
	  (*(tcontext + i )).tid = i;
	  (*(tcontext + i )).stat = READY;
	  (*(tcontext + i )).insocket = -1;
	  (*(tcontext + i )).ServerSockSize = ServerSockSize;
	  (*(tcontext + i )).ClientSockSize = ClientSockSize;
	  (*(tcontext + i )).CopySize = CopySize;
	  if ( sem_init (&(*(tcontext + i )).wait, 0, 0) ) {
	      logmsg (LOG_ERR, "Not enough resources for sem_init().");
	      exit (1);
	  }
      }
  }
  
  /*********** Ende Thread-inits */
  
  /* Threads erzeugen */
  
  if (span_threads (threads, tcontext) < threads ) {
      logmsg(LOG_ERR, "Unable to start requested number of threads.");
  }
  
  /*
   * Open a TCP socket (an Internet stream socket).
   */
  
  if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      logmsg(LOG_ERR, "Unable to open stream socket.");
      exit(1);
  }
  
  /* Socketsize fuer Clientseite fruehzeitig setzen. 
   * wird an sockets, die ueber accept kommen, vererbt.
   * nur so koennen window-sizes richtig ausgehandelt werden.
   */
  if (ClientSockSize)
    setsocksize(sockfd, ClientSockSize);
  
  /*
   * Bind our local address so that the client can send to us.
   */
  setsockreuse(sockfd);

  
  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family      = AF_INET;
  /* hier ansetzen fuer dual-homed-Gateway: */
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port        = htons(serv_port);
  
  if (bind(sockfd, (struct sockaddr *) &serv_addr, 
	   sizeof(struct sockaddr_in)) < 0) {
      logmsg(LOG_ERR, "Unable to bind local address with errno %d.", errno);
      exit(1);
  }
  
  if (listen(sockfd, 5) == -1) {
      logmsg(LOG_ERR, "Listen failed with errno %d.", errno);
      exit(1);
  }
  
  logmsg(LOG_NOTICE, "Starting at port %d.", serv_port);
  
  for ( ; ; ) {
      /*
       * Wait for a connection from a client process.
       * This is an example of a concurrent server.
       */
      clilen = sizeof(cli_addr);
      
      if ((newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr,
			      &clilen)) == -1) {
	  logmsg(LOG_ERR, "Unable to create socket for request connection (accept failed with errno %d).", errno);
	  usleep (500); /* ca.: 500 us warten */
	} else { 
	    /* neuer Socket vom Kernel */
	    int responder;
	    
	    logmsg(LOG_NOTICE, "Created socket for request connection to %s:%d.", 
		   inet_ntoa(cli_addr.sin_addr),
		   cli_addr.sin_port);

	    if (!client_allowed(cli_addr.sin_addr.s_addr)) {
	        logmsg(LOG_NOTICE, "Connect attempt from unknown client.");
		close(newsockfd);
		goto accept_new_sock;
		}
	    
	ll1:
	    responder = find_thread ( threads, tcontext );
	    if (responder >= 0) {
		logmsg(LOG_NOTICE, "Thread #%d is ready to accept connection on socket %d.", responder, newsockfd);

		/* arbeitswilliger Thread gefunden */
		(*(tcontext + responder)).insocket = newsockfd;     /* Socket uebergeben */
		(*(tcontext + responder)).stat = SCHEDULED;       /* markieren fuer find_thread() */
		if (sem_post (&(*(tcontext + responder)).wait)) {  /* freigeben */
		    /* irgendein Fehler ?!, -> thread unbrauchbar */
		    (*(tcontext + responder)).stat = ERROR;
		    logmsg(LOG_ERR,"Thread #%d doesn't accept the semaphore.", responder);
		    goto ll1;                                     /* einen anderen Thread finden */
		} else {
		    logmsg(LOG_NOTICE, "Giving job to thread #%d.", responder);
		}
	    } else if (responder == -1) {
		/* no more unused threads, have to wait */
		usleep (50); /* ca.: 50 us warten */
		goto ll1;
	    } else {
		logmsg (LOG_ERR, "FATAL ERROR: all threads are in ERROR state, aborting.");
		exit (1);
	    }
	}

accept_new_sock:
	/* solaris cc expects at least something here */
	;
  }
}
