/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * $Id: misc.c,v 1.33 2000/05/08 13:11:30 glauschwuffel Exp $
 *
 * $Log: misc.c,v $
 * Revision 1.33  2000/05/08 13:11:30  glauschwuffel
 * another legal_constituent: +
 *
 * Revision 1.32  2000/05/07 15:26:27  glauschwuffel
 * changed inclusion of strings.h to string.h. moved function val_end to parse_tag.c. made function find_unquoted_end recognize the end of tags... new procession stream. function legal_constituent now allows more characters. new func legal_val_constituent.
 *
 * Revision 1.31  2000/04/24 18:00:37  gellert
 * cleaned up the return values to get rid of integer-pointer warnings.
 *
 * Revision 1.30  2000/04/16 15:32:55  gellert
 * output_string called reset_buffers() which could lead to negative string
 * lengths in calls to itself. Fixed and added some checks ;)
 *
 * Revision 1.29  2000/03/07 15:18:01  gellert
 * some changes to port to linux!!!
 *
 * Revision 1.28  2000/03/07 15:06:52  goldbach
 * added GPL info
 *
 * Revision 1.27  2000/03/01 13:33:16  gellert
 * Added a new commandline option "-n" (nodaemon), so the proxy stays
 * in foreground for debugging.
 *
 * Revision 1.26  2000/02/23 15:11:26  goldbach
 * now we use POSIX-semaphores instead of Solaris-semaphores
 *
 * Revision 1.25  2000/02/21 15:08:46  gellert
 * *** empty log message ***
 *
 * Revision 1.24  2000/02/21 14:56:19  gellert
 * Removed the unused function "output_data"
 *
 * Revision 1.23  1999/12/29 17:02:59  gellert
 * added a comma as a legal_constituent for tags and attributes without quotes.
 *
 * Revision 1.22  1999/11/30 16:57:53  gellert
 * skip flag: initialization moved to beginning of copy-pakets. read_header: empty_line is called with the length of the read block, not the full HTTP_BUFSIZE.
 *
 * Revision 1.21  1999/11/26 10:47:07  goldbach
 * New actions introduced: replskip and replendskip skip sections of documents, i.e. data won\'t be sent to the client.
 *
 * Revision 1.20  1999/11/18 12:56:26  gellert
 * fixed a bug in "logmsg". Now the message string is NOT interpreted
 * any more...  syslog(level, logbuf) --> syslog(level, "%s", logbuf)
 *
 * Revision 1.19  1999/11/11 15:51:49  gellert
 * Removed a "flush(stderr" command.
 * Added "_" as allowed char for an attribute.
 *
 * Revision 1.18  1999/08/19 13:43:30  gellert
 * some changes in the handling of quoted values for attributes
 * (now according to html documentation 3.2.2).
 * allowed '?' in unquoted values.
 *
 * Revision 1.17  1999/08/18 12:03:41  gellert
 * gregor fixed the nasty "xxx=''" bug. maybe now quotes really work.
 * removed an unused procedure (a_relevant).
 *
 * Revision 1.16  1999/08/09 13:27:41  goldbach
 * '#' and '%' are now legal constituents, too. this is due to the real world: '#' is used
 * in setting the bakground colour and '%' is used in widths etc. ...
 *
 * Revision 1.15  1999/08/09 12:04:33  goldbach
 * legal_constituent changed to accept '/' and ':' as chars in
 * values without surrounding quotes. This should be still safe.
 *
 * Revision 1.14  1999/07/28 14:22:22  gellert
 * some routines moved to here (legal_constituent etc).
 *
 * Revision 1.13  1999/07/22 14:30:54  gellert
 * some improvements of the determination of the tag end,
 * now nested quotes etc are supported.
 * changed config option replace to replabort/replcont.
 *
 * Revision 1.12  1999/07/16 09:47:55  goldbach
 * minor cleanup (warnings)
 *
 * Revision 1.11  1999/07/15 19:57:47  gellert
 * output_string now does a send if the string does not
 * fit into the outbuf.
 *
 * Revision 1.10  1999/07/14 09:50:05  goldbach
 * removed last strncpy.
 *
 * Revision 1.9  1999/06/30 12:51:33  goldbach
 * removed calls to logmsg() from strnchr
 *
 * Revision 1.8  1999/06/29 12:02:04  gellert
 * new logmsg procedure using vsnprintf.
 *
 * Revision 1.7  1999/06/23 16:58:16  gellert
 * logging is now always done to LOG_INFO, independent of the
 * level given to logmsg (level is just used to determine if it
 * should be logged at all).
 *
 * Revision 1.6  1999/06/23 09:11:01  gellert
 * added initialization of a variable to calm down purify.
 *
 * Revision 1.5  1999/06/09 20:33:54  gellert
 * new version. now the parsing of tags and attributes works.
 * and it still works if the client cancels the transmission
 * (stop button in browser).
 *
 * Revision 1.4  1999/05/27 16:44:15  goldbach
 * added logmsg()
 *
 * Revision 1.3  1999/05/27 11:29:36  gellert
 * well, now it does what we want. just small mistakes (eg end of
 * comments) are done.
 *
 * Revision 1.2  1999/05/12 16:04:21  goldbach
 * major clean-up session
 *
 * Revision 1.1  1999/05/04 09:18:51  goldbach
 * initial version. functions have been in modules where they don't belong.
 * 'tis much cleaner now :)
 *
 */

#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
/* for mkdaemon */
#include <unistd.h>
#include <sys/stat.h>
/* for the logging facility */
#include <syslog.h>
#include <sys/socket.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
/* for the semaphore. vsnprintf is not listed as MT-Safe :( */
/* #include <synch.h> this should not be necessary. evtl. pthread.h */
#include <semaphore.h>

#include <sys/resource.h>

/* and for the thread_context */
#include "mythreads.h"
#include "misc.h"

#include "httpf.h"
#include "parse_tag.h"

/* maximum size of one argument of log */
#define SIZE_ARGBUF 255
/* overall size of log buffer */
#define SIZE_LOGBUF 4096

extern sem_t log_sema;


void 
usage()
{
  printf("Usage:\n"
	 "\thttpf   [-h] [-?] [-p <port>] [-m <copysize>] [-t <threads>]\n"
	 "\t\t[-s <clientside socket size>] [-S <serverside socket size>]\n"
	 "\t\t[-d <destination host> -D <destination port>]\n\n"
	 "Options (default)\n"
	 "\t-h | -?                   Displays this message\n"
	 "\t-v <verbosity>      (5)   Set debug/verbosity level (0-7)\n"
         "\t-p <port>        (1080)   Listen on specified port\n"
         "\t-m <copysize>    (8192)   Maximum amount of data [byte] copied by proxy\n"
         "\t                          during each `read-write'-cycle\n"
	 "\t-n                        No daemon: stay in foreground.\n"
	 "\t-t <threads>       (50)   number of pre-forked threads\n"
	 "\t-s <socket size>    (-)   Socket size [byte] for connection with client\n"
	 "\t-S <socket size>    (-)   Socket size [byte] for connection with server\n"
	 "\t                          System specific default socket sizes may vary\n"
	 "\t-d <host>           (-)   Destination host for `Plug-GW' mode\n"
	 "\t-D <port>           (-)   Port to connect on destination host in `Plug-GW' mode\n"
	 );
}




/*
 * reset_buffers
 *
 * Call this before processing a new stream of data from either
 * the server or the client. Resets the buffers by setting
 * the buffer offsets and lengths to 0.
 */
void reset_buffers(TCT *tct)
{
    tct->inoffset=0;
    tct->inlen=0;
    tct->outoffset=0;
    tct->outlen=0;
}

/*
 * output_string
 *
 * Write string into output buffer and increase the according offset.
 * The output buffer will be sent to the insocket if the string
 * doesn't fit into the remaining space.
 * Returns:
 *   OUTPUT_OK  -- string was written to output buffer. If a send
 *     was necessary it was done.
 *   OUTPUT_FAILED -- the string was too sizy for the output buffer
 *     (length > HTTP_BUFSIZE) of sending the output buffer failed.
 * Notes:
 *   If the string was too sizy for the output buffer it should be
 *   be sent to the insocket in chunks of HTTP_BUFSIZE.
 */
int output_string(char *s, int n, TCT *tct) {
    int restlen = 0;
  
    /* return immediately if we should skip contents */
    if (tct->skip_flag==1) return OUTPUT_OK;
    
    /* return immediately if the string is too sizy for the buffer */
    /* FIXME: output the string in chunks of HTTP_BUFSIZE */
    if (n > HTTP_BUFSIZE) return OUTPUT_FAILED;

    if (n<0) {
	logmsg(LOG_INFO, "[%d] negative size in call of output_string()", tct->tid);
	return OUTPUT_FAILED;
    }

    if (s==NULL) {
	logmsg(LOG_INFO, "[%d] null pointer passed in call of output_string()", tct->tid);
	return OUTPUT_FAILED;
    }


    restlen = HTTP_BUFSIZE - tct->outoffset;

    if (n > restlen) {
	/* this does not fit into the buffer */
	
	logmsg(LOG_DEBUG, "[%d] output buffer full, send needed", tct->tid);
	/* send the outbuf so far */
	if (send(tct->insocket, tct->outbuf, tct->outlen, 0)<0) {
	    logmsg(LOG_INFO, "[%d] sending the output buffer failed", tct->tid);
	    return OUTPUT_FAILED;
	}
	
	tct->outoffset = 0; tct->outlen = 0;
    }
    
    memcpy(tct->outbuf+tct->outoffset, s, n);
    tct->outoffset+=n; tct->outlen=tct->outoffset;

    return OUTPUT_OK;
}

/*
 * error_string
 *
 * Write string into error buffer and increase the according offset.
 */
void error_string(char *s, int n, TCT *tct) 
{
    memcpy(tct->errbuf+tct->erroffset, s, n);
    tct->erroffset+=n;
}


char *strnchr(char *buf, char c, int n)
{
    int i=0;
    
    while (i<n && (buf[i]!=c)) {i++;}
    if (i==n)
	return NULL;
    else
	return buf+i;
}


/*
 * find_eol
 *
 * Find end of line in any http buffer. This should be
 * \r\n... however, some people implement only \n or \r
 * so we have to accept these, too. The following character (-combos)
 * are an EOL: \n, \n\r, \r, \r\n (thus \n\n is still *two*
 * newlines).
 * Inputs:
 *    size -- #of bytes to search in
 *    replace -- if this true we replace both cr and lf
 *           with the null character.
 * Returns: 0 if no eol was found or end of line address.
 * Next character will be the new line.
 */
char *find_eol(char *buf, int size, int replace)
{
    char *cur;
 
    for (cur=buf; cur<buf+size; cur++) {
	/* is it \n or \n\r? */
	if (cur[0]=='\n') {
	    if (replace) cur[0]=0;
	    
	    if (cur[1]=='\r') {
		if (replace) cur[1]=0;
		return cur+1;
	    } else
		return cur;
	}

	
	/* is it \r or \r\n? */
	if (cur[0]=='\r') {
	    if (replace) cur[0]=0;
	    
	    if (cur[1]=='\n') {
		if (replace) cur[1]=0;
		return cur+1;
	    } else
		return cur;
	}
    }
    return 0;
    
}


/* Vom steuernden Terminal unabhaengig machen */
/* von Benecke uebernommen...
 */
int mkdaemon (void)
{
  int pid = fork();
  if (pid > 0)
    exit (0); /* Vaterprozess soll terminieren */
  if (pid == 0)
    {
      setsid();   
      
      chdir ("/"); /* in das Root-Verzeichnis wechseln, so dass wir
                      nicht durch ein UMOUNT der Partition von der wir
                      gestartet wurden, terniniert werden.
                      */
      umask (0);
    }
  return pid;
}





/* #include <stdio.h>
 * #include <sys/types.h>
 */
 
#define		MAXLINE			80

/* gg 29.12.99:
 * move these global variables below into the thread context...
 * These functions are unused for now :)
 */
			
u_long	lAddress=0;
int		wSpace=0;

void htoa(register char bHex, register char *pAsciiRet) {

	if (bHex >= 0 && bHex <= 9) *(pAsciiRet) = bHex + '0';
	else *(pAsciiRet) = (bHex - 10) + 'A';
}

void ctoa(register u_char bChar, register char *pAsciiRet) {
	htoa((bChar>>4),pAsciiRet++);
	htoa((bChar&0x0f),pAsciiRet);
}

void ltoa(register u_long lI, register char *pAsciiRet) {
	char	rgTemp[8];
	register int	I=0;
	register int	J;
	
	while ((lI>>4) != 0) {
		rgTemp[I++] = lI & 0x0f;
		lI = lI>>4;
	}
	rgTemp[I++] = lI;
	for (J=I; J<8; J++) rgTemp[J] = 0;
	while (J > 0) {
		J--;
		htoa(rgTemp[J], pAsciiRet++);
	}
}

void xdline(register char *pLine, register int cLine, FILE *logfd) {
	register int	I, J;
	register int	iCol;
	register int	iAlphaCol;
	char		pAddress[9];
	char		pData[3];
	u_char		Line[MAXLINE];

	for (iCol=0; iCol<MAXLINE; iCol++) Line[iCol] = ' ';
	iCol = 0;
	iAlphaCol = 60;
	ltoa(lAddress,pAddress);
	pAddress[8] = '\0';
	strcpy((char *) &Line[iCol],pAddress);
	iCol += 8;
	Line[iCol++] = ' ';
	Line[iCol++] = ' ';
	lAddress += 0x10;
	for (I=0; I<cLine; I++) {
		if ((*pLine >= ' ') && (*pLine <= '~'))
			Line[iAlphaCol++] = *pLine;
		else Line[iAlphaCol++] = '.';

		ctoa(*(pLine++),pData);
		pData[2] = '\0';
		strcpy((char *) &Line[iCol],pData);
		iCol += 2;
		Line[iCol++] = ' ';
	}
	Line[iAlphaCol++] = '\n';
	for (J=0; J<wSpace; J++)
		Line[iAlphaCol++] = '\n';
	Line[iAlphaCol] = '\0';

	/*
	for (I=0; I<iAlphaCol; I++)
		putchar(Line[I]);
	*/
	fprintf(logfd, "%s", Line);
}

void xdbuf(register char *pBuf, register int cBuf, FILE *logfd) {
	register int	nLine;
	register int	i;

	lAddress = 0;
	nLine = cBuf>>4;
	for (i=0; i<nLine; i++) {
		xdline(pBuf, 0x10, logfd);
		pBuf += 0x10;
	}
	i = cBuf & 0x0f;
	if (i > 0) {
		xdline(pBuf, i, logfd);
	}
}

	 
/*
 * logmsg
 *
 * Log information (via syslog()) if given logging level is
 * not higher than the global level (LOG_DEBUG is high).
 * Uses vsnprintf() and a semaphore.
 *
 * Inputs:
 *   log_level -- passed to syslog().
 *   fmt -- printf-like format string.
 *   ... -- args needed by format string
 */

int logmsg (int level, const char *fmt, ...) 
{
    va_list ap;
    char logbuf[SIZE_LOGBUF];
    int result;
    
    /* don't log if level is too high */
    if (level > log_level) return 0;

    va_start(ap, fmt);

    sem_wait(&log_sema); /* wait for semaphore */
    /* format string */
    result = vsnprintf(logbuf, SIZE_LOGBUF, fmt, ap);
    sem_post(&log_sema);
    
    /* this prevents the content of logbuf to be interpreted */
    syslog(level, "%s", logbuf);

    if (result>SIZE_LOGBUF) {
      syslog(level, " log_msg: message was truncated");
    }
      
    va_end(ap);

    return 0;
}


int test_write(int the_socket) {

  int 	nfds = getdtablesize(); /* select ...? */
  fd_set 	fds;	/* fuer select: liste mit sockin + sockout */
  static struct timeval	tout;
  
  FD_ZERO(&fds);
  
      tout.tv_sec = 60*60*2;
      tout.tv_usec = 0;

      FD_SET(the_socket, &fds);

      if (select(nfds, NULL, &fds, NULL, &tout) > 0) {
	  /* writeable? */
	  if (FD_ISSET(the_socket, &fds)) {
	      /* fprintf(stderr, "test_write: yes!\n"); */
	      return 1;
	  } else {
	      /* fprintf(stderr, "test_write: no!\n"); */
	      return 0;
	  }
      }
      
      /* fprintf(stderr, "test_write: no!\n"); */
      return 0;
}



/*
 * a_quote
 *
 * Is there a [HTML-]quote? If so, return the ASCII-quote.
 *
*/
char a_quote(char *string) {
    if (strncmp(string, "\"", 1)==0)
	return string[0];
        
    if (strncmp(string, "\'", 1)==0)
	return string[0];
        
    if (strncmp(string, "&quot;", 6)==0)
      return '\"';
    
    if (strncmp(string, "&#39;", 5)==0)
      return '\'';
    
    if (strncmp(string, "&#34;", 5)==0)
	return '\"';
    
    return 0;
}

/*
 * skip_quote
 *
 * Take address where a quote starts and return the address
 * where it ends. We don't check if there's quote.
 */
char *skip_quote (char *s) {
  if (s[0]=='&') {
	if (s[1]=='q')
	  /* &quot; */
	  return s+6;
	else
	  /* &#34; and &#39; */
	  return s+5;
  } else {
	/* single and double quote */
	return s+1;
  }
}


/*
 * find_end
 *
 * Find the end of a correctly nested string.
 * TRAILER is the char the string has to end with.
 * Gets address of the last character (the trailer).
 * Returns:
 *  NULL - invalid attribute. Quotes not nested properly so
 *    we reached the end of the input buffer.
 *  other - addr. of last char.
 * Note: This would be the place to insert a check for the
 *   maximum length of strings.
 */
char *find_end(char *start, char trailer, TCT *tct) {
  char *pos;
  char c;
    
  pos = start;

  logmsg(LOG_DEBUG, "[%d] Next level: find %c from %c%c%c....",
		 tct->tid, trailer, start[0], start[1], start[2]); 
    
  /* as long as we are in the string */
  while((pos[0]!=0) && (pos < tct->inbuf+tct->inlen)) {
	c = a_quote(pos); /* quote? */
	if (c!=0) {
	  /* we have a quote */
	  switch(trailer) {
	  case '\"':
		if (c=='\'') {
		  logmsg(LOG_DEBUG, "[%d] skip single",tct->tid);
		  pos+=1; /* skip it */
		} else {
		  if (c=='\"') {
			/* match! */
			return skip_quote(pos);
		  }
		}
		break;
		
	  case '\'':
		if (c=='\"') {
		  pos+=1; /* skip it */
		} else {
		    if (c=='\'') {
			  /* match! */
			  return skip_quote(pos);
		    }
		}
		break;
	  }
	} else {
	  /* we don't have a quote */
	  
	  /* gg 18.08.99: proceed in the buffer only if we didn't encounter
		 a quote. Otherwise things like 
		 ... window.status=''">
		 won't be parsed correctly since we happily skip the closing double quote.
		 Bah. This was a naughty bug. :( */
	  pos++;
	}
  }
  
  /* if we are here we didn't find a matching quote. */
  return NULL;
}


/*
 * find_unquoted_end
 *
 * Find end of unquoted value. This should be either a blank
 * or the end of the string.
 * Returns:
 *   address behind the last character of the value.
 *   0 if the value is illegal
 *
 * Note: the HTML 4.0 spec says an unquoted value may consist of
 * decimals, letters, hyphens and periods.
 * (3.2.2 Attributes HTML 4.0 Specification)
 */
char *find_unquoted_end(char *val_start) {
  int i=0;

  /* skip legal characters */
  while(legal_val_constituent(val_start[i])) {
	i+=1;
  }

  /* whitespace and '>' are okay */
  if (legal_whitespace(val_start[i]) || val_start[i]=='>')
	return val_start+i;
  else
	/* illegal character encountered, error */
	return NULL;
}

/*
 * legal_constituent
 *
 * Takes character and tests if this is a legal constituent
 * for a tag or unquoted attribute, i.e. if it's a letter, digit,
 * period or hyphen.
 * Due to the real world (e.g. yahoo) we allow the slash '/' and the
 * colon ':', too. And '#' (setting colours), '%' (widths etc.) and '?'
 * (cgi), ',' (in frameset: "ROWS=20%,50%,30%" :|).
 * Returns character if legal, 0 otherwise.
 * Note: since we allow '&' we MUST NOT allow '#'... if we did we would
 * allow character references...
 * Do not allow parens... these are usually used in code :)XS
 */
int legal_constituent(int c) {
  if (isalnum(c) || c=='.' || c=='-' || c=='/' || c==':' || c=='&' || c=='%' || c=='_' || c=='?' || c==',' || c=='@' || c=='~' || c=='+')
	return c;
  else
	return 0;
}

int legal_val_constituent(int c) {
  if (legal_constituent(c) || c=='=')
	return c;
  else
	return 0;
}

/*
 * legal_whitespace
 *
 * Takes character and tests if this is a legal whitespace.
 * Basically it's a isspace() wrapper ;)
 * Returns character if legal, 0 otherwise.
 */
int legal_whitespace(int c)
{
    if (isspace(c))
	return c;
    else
	return 0;
}


/*
 * skip_leading_whitespace
 *
 * Skip legal whitespace.
 */
char *skip_leading_whitespace(char *buf) {
  int i=0;
  
  /* skip leading blanks while we are in the string*/
  while(buf[i] && legal_whitespace(buf[i])) {
	i+=1;
  }
  return buf+i;
}

/*
 * skip_to_next_whitespace()
 *
 * Proceed to the next whitespace character.
 * Returns pointer to whitespace character or
 * NULL if the string ends.
 */
char *skip_to_next_whitespace(char *buf) {
  int i=0;

  if (buf==NULL) return NULL;

  /* skip characters while we are in the string*/
  while(buf[i] && !legal_whitespace(buf[i])) {
	i+=1;
  }
  if (buf[i]==0) return NULL;

  return buf+i;
}
